/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */
package org.ow2.weblab.petals;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.Fault;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;

import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.weblab.petals.fault.ExceptionFaultMapper;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * In charge of converting web services methods to JBI calls.
 * 
 * @author Jérémie Doucy, CASSIDIAN, an EADS company
 * 
 */
public class JbiInvocationHandler implements InvocationHandler {

	protected ComponentContext ctx;
	protected Logger logger;
	protected QName serviceName;
	protected QName interfaceName;

	private static final Map<Method, JAXBContext> JC_MAP = Collections.synchronizedMap(new HashMap<Method, JAXBContext>());
	private static final Map<Method, ExceptionFaultMapper> EXP_MAP = Collections.synchronizedMap(new HashMap<Method, ExceptionFaultMapper>());

	/**
	 * Initialize invocation handler with needed JBI parameters.
	 * 
	 * @param interfaceName
	 *            could be null, if not specify the service interface of the endpoint to call (if there is multiple interfaces for the same serviceName)
	 * @param serviceName
	 *            serviceName used in the bus for this call
	 * @param ctx
	 *            bus context
	 * @param logger
	 *            to log
	 */
	public JbiInvocationHandler(final QName interfaceName, final QName serviceName, final ComponentContext ctx, final Logger logger) {
		super();
		this.interfaceName = interfaceName;
		this.serviceName = serviceName;
		this.logger = logger;
		this.ctx = ctx;
	}

	@Override
	public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {

		this.logger.info("Calling method: " + method.getName());
		/*
		 * only ones
		 */
		synchronized (JbiInvocationHandler.JC_MAP) {
			if (!JbiInvocationHandler.JC_MAP.containsKey(method)) {
				/*
				 * initialize marshaler with all needed classes (using annotation on Method).
				 */
				JbiInvocationHandler.JC_MAP.put(method, this.initJC(method));
				/*
				 * initialize SOAP Fault -> Java Exception mapping for this method (also uses annotation on Method object)
				 */
				JbiInvocationHandler.EXP_MAP.put(method, new ExceptionFaultMapper(method, this.logger));
			}
		}

		/*
		 * creates the exchange for this call, using context
		 */
		final ExchangeImpl exchange = new ExchangeImpl(this.ctx.getDeliveryChannel().createExchangeFactory()
				.createExchange(MEPConstants.IN_OUT_PATTERN.value()));

		/*
		 * set operation and service to call
		 */
		exchange.setOperation(this.getOperation(method));
		exchange.setService(this.serviceName);

		if (this.interfaceName != null) {
			this.logger.info("Using interface: " + this.interfaceName + " extracted from " + WebService.class + " annotation.");
			exchange.setInterfaceName(this.interfaceName);
		}
		/*
		 * transform Java JAXB parameter to Source XML message to put inside inMessage of exchange
		 */
		try {
			exchange.setInMessageContent(this.generateInMessageContent(args[0], method));

		} catch (final JAXBException jaxbException) {
			jaxbException.printStackTrace();
		}
		this.logger.info("Calling: " + this.serviceName);
		/*
		 * send exchange to endpoint
		 */
		this.ctx.getDeliveryChannel().sendSync(exchange.getMessageExchange());
		this.logger.info("Get reply from: " + this.serviceName);
		if (exchange.isErrorStatus()) {
			this.logger.info("Error: " + exchange.getError());
		} else if (exchange.getFault() != null) {
			/*
			 * fault
			 */
			final Fault fault = exchange.getFault();
			final DOMResult domResult = new DOMResult();
			TransformerFactory.newInstance().newTransformer().transform(fault.getContent(), domResult);
			this.logger.fine("Fault in message, trying to map it to an exception.");

			/*
			 * uses this Method fault exception mapping to generate Java Exception.
			 */
			final Exception exception = JbiInvocationHandler.EXP_MAP.get(method).generateException(domResult.getNode());
			this.logger.info("Fault in message, throwing: " + exception.getClass());
			throw exception;

		}
		return this.generateOutRet(exchange.getOutMessageContentAsDocument(), method);

	}

	private QName getOperation(final Method method) {
		return new QName(method.getName());
	}

	/**
	 * Marshal a Java object using JAXB to a DOM document.
	 * 
	 * @param arg
	 *            the Java object to marshal
	 * @param method
	 *            the method used to get name space of the generated DOM Document using WebParam annotation
	 * @return the marshaled DOM Document
	 * @throws JAXBException
	 *             Something wrong during parameter marshaling
	 */
	private Document generateInMessageContent(final Object arg, final Method method) throws JAXBException {
		/*
		 * instanciate marshaler using JaxbContext map
		 */
		final Marshaller marshaller = JbiInvocationHandler.JC_MAP.get(method).createMarshaller();
		final WebParam webParam = (WebParam) method.getParameterAnnotations()[0][0];

		/*
		 * marshall to DOM Document
		 */
		final JAXBElement<Object> jaxbElement = new JAXBElement<Object>(new QName(webParam.targetNamespace(), webParam.name()), Object.class, arg);
		final DOMResult domResult = new DOMResult();
		marshaller.marshal(jaxbElement, domResult);

		return (Document) domResult.getNode();
	}

	/**
	 * Unmarshal DOM Node to a JAXB annotated Java object using Method return type to precise the class of this Java object.
	 * 
	 * @param node
	 *            DOM Node to unmarshal
	 * @param method
	 *            used to get return type
	 * @return the Java Object corresponding to the return type
	 * @throws JAXBException
	 *             Something wrong during unmarshaling
	 */
	private Object generateOutRet(final Node node, final Method method) throws JAXBException {
		/*
		 * instanciate unmarshaler using JaxbContext map
		 */
		final Unmarshaller unmarshaller = JbiInvocationHandler.JC_MAP.get(method).createUnmarshaller();
		/*
		 * unmarshal to return type Java Object
		 */
		final Object ret = unmarshaller.unmarshal(node, method.getReturnType()).getValue();
		return ret;
	}

	/**
	 * Initialize a JaxbContext with method parameters.
	 * 
	 * @param method
	 *            used to find parameters
	 * @return Initialized JaxbContext
	 */
	protected JAXBContext initJC(final Method method) {
		try {
			final List<Class<?>> classes = new ArrayList<Class<?>>();
			for (final Class<?> clazz : method.getParameterTypes()) {
				if (!classes.contains(clazz)) {
					classes.add(clazz);
				}
			}
			if (!classes.contains(method.getReturnType())) {
				classes.add(method.getReturnType());
			}
			final Class<?>[] classesTab = new Class<?>[classes.size()];
			int i = 0;
			for (final Class<?> clazz : classes) {
				classesTab[i] = clazz;
				i++;
			}

			return JAXBContext.newInstance(classesTab);
		} catch (final JAXBException e) {
			e.printStackTrace();
			throw new RuntimeException("GRAVE: Error during marshaller instanciation.", e);
		} catch (final Exception e) {
			e.printStackTrace();
			throw new RuntimeException("GRAVE: Error during marshaller instanciation.", e);
		}
	}

}
