/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */
package org.ow2.weblab.petals;

import java.lang.reflect.Proxy;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jws.WebService;
import javax.xml.namespace.QName;

/**
 * This class uses the Proxy pattern (@see java.lang.reflect.Proxy) to dynamically instanciate a Proxy of the parameterized interface type.
 * 
 * @author Jérémie Doucy, CASSIDIAN, an EADS company
 * 
 */
public class JbiServiceFactory {

	/**
	 * Generate a JBI Proxy service which call a JBI endpoint when invoked. Use this method if you want to force an interface name (if there is multiple
	 * interfaces for the same serviceName). If this interface name is null, only the serviceName will be used for endpoint resolving (could be dangerous in
	 * case of multiple interface for the same service name). If the interfaceName is mentioned in the @WebService annotation on the specifiediInterface class,
	 * you can use the simplest method (whithout specifying an interface name):
	 * 
	 * @see getProxy(final Class<ServiceInterface>, final QName, final ComponentContext, final Logger)
	 * 
	 * @param <ServiceInterface>
	 *            interface of the returned object (the service class)
	 * @param clazz
	 *            the class of the interface
	 * @param interfaceName
	 *            the service interface of the endpoint to call
	 * @param serviceName
	 *            the name of the service in the bus (to retrieve the endpoint to call)
	 * @param ctx
	 *            bus context used to create and send message exchange
	 * @param logger
	 *            the logger to log ...
	 * @return a Proxy which is an instance of <ServiceInterface>, each WebMethod defined on this interface will be delegated to a generic call inside the bus
	 */
	@SuppressWarnings("unchecked")
	public static <ServiceInterface> ServiceInterface getProxy(final Class<ServiceInterface> clazz, final QName interfaceName, final QName serviceName,
			final ComponentContext ctx, final Logger logger) {
		return (ServiceInterface) Proxy.newProxyInstance(clazz.getClassLoader(), new Class[] { clazz }, new JbiInvocationHandler(interfaceName, serviceName,
				ctx, logger));
	}

	/**
	 * Generate a JBI Proxy service which call a JBI endpoint when invoked. Use this method if you don't want to specify a specific service interface. This
	 * method reads the annotation @WebService on the specified interface Class to generate the interface name used in the bus to reach the good endpoint.
	 * 
	 * @param <ServiceInterface>
	 *            interface of the returned object (the service class)
	 * @param clazz
	 *            the class of the interface
	 * @param serviceName
	 *            the name of the service in the bus (to retrieve the endpoint to call)
	 * @param ctx
	 *            bus context used to create and send message exchange
	 * @param logger
	 *            the logger to log ...
	 * @return a Proxy which is an instance of <ServiceInterface>, each WebMethod defined on this interface will be delegated to a generic call inside the bus
	 */
	@SuppressWarnings("unchecked")
	public static <ServiceInterface> ServiceInterface getProxy(final Class<ServiceInterface> clazz, final QName serviceName, final ComponentContext ctx,
			final Logger logger) {
		/*
		 * searching for WebService annotation on one of the proxy interface to put service interface
		 */
		QName interfaceName = null;
		logger.warning("Searching for WebService annotation on interface: " + clazz + " to define interface name used in the bus.");
		if (clazz.isAnnotationPresent(WebService.class)) {
			WebService webServiceAnnot = clazz.getAnnotation(WebService.class);
			interfaceName = new QName(webServiceAnnot.targetNamespace(), webServiceAnnot.name());
		} else {
			logger.warning("No WebService annotation found on interface: " + clazz + ". Only the serviceName will be used to find the right endpoint.");
		}
		return (ServiceInterface) Proxy.newProxyInstance(clazz.getClassLoader(), new Class[] { clazz }, new JbiInvocationHandler(interfaceName, serviceName,
				ctx, logger));
	}
}
