/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2012 CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */
package org.ow2.weblab.petals.fault;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import javax.xml.namespace.QName;
import javax.xml.ws.WebFault;

import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * Enable exception generation from a DOM node corresponding to a SOAP 1.1 Fault. Uses Method reflection and WebFault annotation reading to map XML to
 * Exception. You must use one instance of this class per Method.
 * 
 * @author Jérémie Doucy, CASSIDIAN, an EADS company
 * 
 */
public class ExceptionFaultMapper {

	/*
	 * the logger
	 */
	private Logger logger;

	/*
	 * QName detail fault value -> Java exception class
	 */
	private Map<QName, Class<?>> exceptionMap;

	/**
	 * Populate the exception map from a Method reflection object using WebFault annotation. Only classes with WebFault annotation are added.
	 * 
	 * @param method
	 *            the method
	 */
	public ExceptionFaultMapper(final Method method, final Logger logger) {
		super();
		this.logger = logger;
		this.exceptionMap = new HashMap<QName, Class<?>>();
		this.logger.info("Extracting WebFault from method: " + method.getName());
		for (Class<?> clazz : method.getExceptionTypes()) {
			WebFault webFault = clazz.getAnnotation(WebFault.class);
			if (webFault == null) {
				logger.warning("Exception: " + clazz + " not added because no WebFault annotation found.");
			} else {
				QName faultQName = new QName(webFault.targetNamespace(), webFault.name());
				this.exceptionMap.put(faultQName, clazz);
				this.logger.info("Mapping " + clazz + " to " + faultQName);
			}
		}
		/*
		 * no exception with webfault on this method, just warn
		 */
		if (this.exceptionMap.size() == 0) {
			this.logger.warning("No exception detected on this method: " + method.getName() + ". Nothing to map.");
		}

	}

	/**
	 * Parse SOAP 1.1 Fault and instanciate an exception with faultstring generic SOAP Fault value and specific value from specific exception. Exception
	 * constructor must manage two String.
	 * 
	 * @param node
	 * @return
	 */
	public Exception generateException(Node node) {
		Exception ret = new Exception("Unmappable exception");
		/*
		 * search DOM node corresponding to detail value (SOAP 1.1)
		 */
		Node details = findDetails(node);
		/*
		 * search standard faultstring (SOAP 1.1)
		 */
		String faultString = findFaultString(node);
		Class<Exception> exClass = null;
		/*
		 * search for node with element corresponding to one of the QName->Exception map
		 */
		int i = 0;
		while (i < details.getChildNodes().getLength() && exClass == null) {
			Node cur = details.getChildNodes().item(i);
			if (cur.getLocalName() != null && cur.getNamespaceURI() != null) {
				QName qName = new QName(cur.getNamespaceURI(), cur.getLocalName());
				if (this.exceptionMap.containsKey(qName)) {
					/*
					 * QName found, try to instanciate the corresponding class
					 */
					Class<?> clazz = null;
					try {
						clazz = this.exceptionMap.get(qName);
						return (Exception) clazz.getConstructor(String.class, String.class).newInstance(faultString, getExceptionValue(cur));

					} catch (Exception e) {
						e.printStackTrace();
						logger.warning("Error during " + clazz + " instanciation: " + e.getMessage());
					}
				}
			}
			i++;

		}
		return ret;
	}

	/**
	 * Return string value of detail exception node
	 * 
	 * @param detailNode
	 *            SOAP 1.1 detail node
	 * @return the extracted String value
	 */
	private String getExceptionValue(Node detailNode) {
		String value = null;
		int i = 0;
		while (i < detailNode.getChildNodes().getLength() && value == null) {
			Node cur = detailNode.getChildNodes().item(i);
			if (cur instanceof Text) {
				Text text = (Text) cur;
				value = text.getTextContent();
			}
			i++;
		}
		return value;
	}

	/**
	 * Recursive method to find fault string value.
	 * 
	 * @param node
	 *            the to search in
	 * @return fault string value
	 */
	private String findFaultString(Node node) {
		if (node.getNodeName().equalsIgnoreCase("faultstring")) {
			return node.getTextContent().trim();
		} else if (node.hasChildNodes()) {
			int i = 0;
			while (i < node.getChildNodes().getLength()) {
				String value = findFaultString(node.getChildNodes().item(i));
				if (value != null) {
					return value;
				}
				i++;

			}
		}
		return null;
	}

	/**
	 * Recursive method to find detail node.
	 * 
	 * @param node
	 *            the to search in
	 * @return detail node
	 */

	private Node findDetails(Node node) {
		if (node.getNodeName().equalsIgnoreCase("detail")) {
			return node;
		} else if (node.hasChildNodes()) {
			int i = 0;
			while (i < node.getChildNodes().getLength()) {
				Node details = findDetails(node.getChildNodes().item(i));
				if (details != null) {
					return details;
				}
				i++;

			}
		}
		return null;
	}
}